package ru.dbotthepony.mc.otm.core

import com.google.common.collect.ImmutableList
import net.minecraft.network.chat.Component
import java.math.BigDecimal
import java.math.BigInteger

fun BigInteger.formatReadableNumber(): String {
	if (isZero) {
		return "0"
	}

	val strValue = toString()

	val absLength = strValue.length - (if (isNegative) 1 else 0)
	val remainder = absLength % 3
	var groups = absLength / 3

	if (remainder == 0) {
		groups--
	}

	val buffer = CharArray((if (remainder == 0) 3 else remainder) + groups * 4 + if (isNegative) 1 else 0)
	var c = 0
	var index = buffer.size - 1

	for (i in strValue.length - 1 downTo (if (isNegative) 1 else 0)) {
		c++

		if (c == 4) {
			buffer[index] = ' '
			index--
			c = 1
		}

		buffer[index] = strValue[i]
		index--
	}

	if (isNegative) {
		buffer[index] = '-'
	}

	return String(buffer)
}

enum class SiPrefix(
	val power: Int,
	fractional: Boolean,
	val symbol: Char,
) {
	// multiplies
	KILO (3,  false, 'k'),
	MEGA (6,  false, 'M'),
	GIGA (9,  false, 'G'),
	TERA (12, false, 'T'),
	PETA (15, false, 'P'),
	EXA  (18, false, 'E'),
	ZETTA(21, false, 'Z'),
	YOTTA(24, false, 'Y'),

	// decimals
	DECI (1,  true,  'd'),
	CENTI(2,  true,  'c'),
	MILLI(3,  true,  'm'),
	MICRO(6,  true,  'μ'),
	NANO (9,  true,  'n'),
	PICO (12, true,  'p'),
	FEMTO(15, true,  'f'),
	ATTO (18, true,  'a'),
	ZEPTO(21, true,  'z'),
	YOCTO(24, true,  'y');

	val formatLocaleKey = "otm.suffix.${name.lowercase()}".intern()
	val rawLocaleKey = "otm.suffix_raw.${name.lowercase()}".intern()

	val string = if (fractional) "0." + "0".repeat(power - 1) + "1" else "1" + "0".repeat(power)

	fun paddedIndex(input: String, index: Int): Char {
		val finalIndex = input.length - power + index

		if (finalIndex >= 0) {
			return input[finalIndex]
		}

		return '0'
	}

	val decimal = BigDecimal(string)
	val fraction = Fraction(decimal)
	val impreciseFraction = ImpreciseFraction(string)
	val integer = if (!fractional) BigInteger(string) else null

	companion object {
		val MULTIPLIES: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			.add(KILO)
			.add(MEGA)
			.add(GIGA)
			.add(TERA)
			.add(PETA)
			.add(EXA)
			.add(ZETTA)
			.add(YOTTA)
			.build()

		val DECIMALS: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			.add(DECI)
			.add(CENTI)
			.add(MILLI)
			.add(MICRO)
			.add(NANO)
			.add(PICO)
			.add(FEMTO)
			.add(ATTO)
			.add(ZEPTO)
			.add(YOCTO)
			.build()

		val DECIMALS_IMPRECISE: List<SiPrefix> = ImmutableList.builder<SiPrefix>()
			.add(DECI)
			.add(CENTI)
			.add(MILLI)
			.add(MICRO)
			.build()
	}
}

fun BigDecimal.determineSiPrefix(): SiPrefix? {
	if (isZero) {
		return null
	}

	var num = this

	if (isNegative) {
		num = -this
	}

	var prev: SiPrefix? = null

	if (num >= BigDecimal.ONE) {
		for (value in SiPrefix.MULTIPLIES) {
			if (value.decimal <= num) {
				prev = value
			} else {
				break
			}
		}
	} else {
		for (value in SiPrefix.DECIMALS) {
			if (value.decimal >= num) {
				prev = value
			} else {
				break
			}
		}
	}

	return prev
}

fun BigDecimal.formatSiTranslatable(): Component {
	if (isZero) {
		return TextComponent("0.00")
	} else if (this == BigDecimal.ONE) {
		return TextComponent("1.00")
	}

	return TextComponent("1.00")
}

fun BigInteger.determineSiPrefix(): SiPrefix? {
	if (isZero) {
		return null
	}

	var num = this

	if (isNegative) {
		num = -this
	}

	var prev: SiPrefix? = null

	if (num >= BigInteger.ONE) {
		for (value in SiPrefix.MULTIPLIES) {
			if (value.integer!! <= num) {
				prev = value
			} else {
				break
			}
		}
	}

	return prev
}

fun BigInteger.formatSi(decimalPlaces: Int = 2): String {
	require(decimalPlaces >= 0) { "Invalid amount of decimal places required: $decimalPlaces" }
	val prefix = determineSiPrefix() ?: return toString() // + "." + "0".repeat(decimalPlaces)
	val isNegative = isNegative
	val arr = (if (isNegative) -this else this).divideAndRemainder(prefix.integer)
	val divided = arr[0].toString()
	val remainder = arr[1].toString()

	if (decimalPlaces == 0) {
		if (isNegative) {
			return "-" + divided + prefix.symbol
		} else {
			return divided + prefix.symbol
		}
	}

	@Suppress("NAME_SHADOWING")
	val decimalPlaces = decimalPlaces.coerceAtMost(prefix.power)

	val add = (if (isNegative) 1 else 0)

	val buffer = CharArray(divided.length + 2 + decimalPlaces + add)
	buffer[buffer.size - 1] = prefix.symbol

	if (isNegative) {
		buffer[0] = '-'
	}

	for (i in divided.indices) {
		buffer[add + i] = divided[i]
	}

	buffer[add + divided.length] = '.'

	for (i in 0 until decimalPlaces) {
		buffer[add + i + divided.length + 1] = prefix.paddedIndex(remainder, i)
	}

	return String(buffer)
}

fun BigInteger.formatSiComponent(suffix: Any = "", decimalPlaces: Int = 2): Component {
	require(decimalPlaces >= 0) { "Invalid amount of decimal places required: $decimalPlaces" }
	val prefix = determineSiPrefix() ?: return if (suffix == "") TextComponent(toString(decimalPlaces)) else if (suffix is Component) TextComponent(toString(decimalPlaces) + suffix.string) else TextComponent(toString(decimalPlaces) + suffix)
	val isNegative = isNegative
	val arr = (if (isNegative) -this else this).divideAndRemainder(prefix.integer)
	val divided = arr[0].toString()
	val remainder = arr[1].toString()

	if (decimalPlaces == 0) {
		if (isNegative) {
			return TranslatableComponent(prefix.formatLocaleKey, "-$divided", suffix)
		} else {
			return TranslatableComponent(prefix.formatLocaleKey, divided + prefix.symbol, suffix)
		}
	}

	@Suppress("NAME_SHADOWING")
	val decimalPlaces = decimalPlaces.coerceAtMost(prefix.power)

	val add = (if (isNegative) 1 else 0)

	val buffer = CharArray(divided.length + 1 + decimalPlaces + add)

	if (isNegative) {
		buffer[0] = '-'
	}

	for (i in divided.indices) {
		buffer[add + i] = divided[i]
	}

	buffer[add + divided.length] = '.'

	for (i in 0 until decimalPlaces) {
		buffer[add + i + divided.length + 1] = prefix.paddedIndex(remainder, i)
	}

	return TranslatableComponent(prefix.formatLocaleKey, String(buffer), suffix)
}

fun ImpreciseFraction.determineSiPrefix(): SiPrefix? {
	if (isZero) {
		return null
	}

	var num = this

	if (isNegative) {
		num = -this
	}

	var prev: SiPrefix? = null

	if (num >= ImpreciseFraction.ONE) {
		for (value in SiPrefix.MULTIPLIES) {
			if (value.impreciseFraction <= num) {
				prev = value
			} else {
				break
			}
		}
	} else {
		for (value in SiPrefix.DECIMALS_IMPRECISE) {
			if (value.impreciseFraction >= num) {
				prev = value
			} else {
				break
			}
		}
	}

	return prev
}

fun ImpreciseFraction.formatSi(decimalPlaces: Int = 2): String {
	require(decimalPlaces >= 0) { "Invalid amount of decimal places required: $decimalPlaces" }
	val prefix = determineSiPrefix() ?: return toString(decimalPlaces)
	return (this / prefix.impreciseFraction).toString(decimalPlaces) + prefix.symbol
}

fun ImpreciseFraction.formatSiComponent(suffix: Any = "", decimalPlaces: Int = 2): Component {
	require(decimalPlaces >= 0) { "Invalid amount of decimal places required: $decimalPlaces" }
	val prefix = determineSiPrefix() ?: return if (suffix == "") TextComponent(toString(decimalPlaces)) else if (suffix is Component) TextComponent(toString(decimalPlaces) + " " + suffix.string) else TextComponent(toString(decimalPlaces) + " " + suffix)
	return TranslatableComponent(prefix.formatLocaleKey, (this / prefix.impreciseFraction).toString(decimalPlaces), suffix)
}

val POWER_NAME = TranslatableComponent("otm.gui.power.name")
val MATTER_NAME = TranslatableComponent("otm.gui.matter.name")

fun ImpreciseFraction.formatPower(decimalPlaces: Int = 2) = formatSiComponent(POWER_NAME, decimalPlaces)
fun ImpreciseFraction.formatMatter(decimalPlaces: Int = 2) = formatSiComponent(MATTER_NAME, decimalPlaces)
fun ImpreciseFraction.formatMatterFull(decimalPlaces: Int = 2) = TranslatableComponent("otm.gui.matter.format", formatSiComponent(MATTER_NAME, decimalPlaces))

fun BigInteger.formatPower(decimalPlaces: Int = 2) = formatSiComponent(POWER_NAME, decimalPlaces)
fun BigInteger.formatMatter(decimalPlaces: Int = 2) = formatSiComponent(MATTER_NAME, decimalPlaces)
fun BigInteger.formatMatterFull(decimalPlaces: Int = 2) = TranslatableComponent("otm.gui.matter.format", formatSiComponent(MATTER_NAME, decimalPlaces))

fun formatPowerLevel(a: ImpreciseFraction, b: ImpreciseFraction, decimalPlaces: Int = 2) = TranslatableComponent("otm.gui.level", a.formatPower(decimalPlaces), b.formatPower(decimalPlaces))
fun formatMatterLevel(a: ImpreciseFraction, b: ImpreciseFraction, decimalPlaces: Int = 2) = TranslatableComponent("otm.gui.level", a.formatMatter(decimalPlaces), b.formatMatter(decimalPlaces))
