package ru.dbotthepony.mc.otm.data

import com.google.common.collect.ImmutableList
import com.google.gson.*
import com.mojang.serialization.Codec
import com.mojang.serialization.DataResult
import com.mojang.serialization.Dynamic
import com.mojang.serialization.JsonOps
import com.mojang.serialization.codecs.RecordCodecBuilder
import it.unimi.dsi.fastutil.objects.ObjectArrayList
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.item.ItemStack
import net.minecraft.world.level.storage.loot.Deserializers
import net.minecraft.world.level.storage.loot.LootContext
import net.minecraft.world.level.storage.loot.LootPool
import net.minecraft.world.level.storage.loot.predicates.LootItemCondition
import net.minecraftforge.common.loot.IGlobalLootModifier
import net.minecraftforge.common.loot.LootModifier
import net.minecraftforge.registries.ForgeRegistries
import ru.dbotthepony.mc.otm.core.registryName

// 1.19 do be like overengineering already overengineered stuff beyond any recognition
private fun getJson(it: Dynamic<Any?>): JsonElement {
	val value = it.value ?: throw NullPointerException("value was null")
	return value as? JsonElement ?: it.ops.convertTo(JsonOps.INSTANCE, it)
}

private val LootPoolCodec: Codec<Array<LootPool>> by lazy {
	val serializer = Deserializers.createLootTableSerializer().create()

	Codec.PASSTHROUGH.flatXmap({
		try {
			DataResult.success(serializer.fromJson(getJson(it as Dynamic<Any?>), Array<LootPool>::class.java))
		} catch(err: JsonSyntaxException) {
			DataResult.error(err.message)
		}
	}, {
		try {
			DataResult.success(Dynamic(JsonOps.INSTANCE, serializer.toJsonTree(it)))
		} catch(err: JsonSyntaxException) {
			DataResult.error(err.message)
		}
	})
}

class LootTableAppender(conditions: Array<out LootItemCondition>, private vararg val pools: LootPool) : LootModifier(conditions) {
	override fun doApply(generatedLoot: ObjectArrayList<ItemStack>, context: LootContext): ObjectArrayList<ItemStack> {
		pools.forEach { it.addRandomItems(generatedLoot::add, context) }
		return generatedLoot
	}

	override fun codec(): Codec<out IGlobalLootModifier> {
		return CODEC
	}

	companion object {
		val CODEC: Codec<LootTableAppender> by lazy {
			RecordCodecBuilder.create {
				codecStart(it).and(
					LootPoolCodec.fieldOf("pools").forGetter { appender -> appender.pools as Array<LootPool> }
				).apply(it, ::LootTableAppender)
			}
		}
	}
}

class LootTableBasicAppender(conditions: Array<out LootItemCondition>, entries: List<Pair<ItemStack, Double>>) : LootModifier(conditions) {
	constructor(conditions: Array<out LootItemCondition>, vararg entries: Pair<ItemStack, Double>) : this(conditions, entries.toList())

	private val entries = ImmutableList.copyOf(entries)

	override fun doApply(generatedLoot: ObjectArrayList<ItemStack>, context: LootContext): ObjectArrayList<ItemStack> {
		for ((item, chance) in entries) {
			if (context.random.nextDouble() <= chance) {
				generatedLoot.add(item.copy())
			}
		}

		return generatedLoot
	}

	override fun codec(): Codec<out IGlobalLootModifier> {
		return CODEC
	}

	companion object {
		val ItemPairListCodec: Codec<List<Pair<ItemStack, Double>>> by lazy {
			Codec.PASSTHROUGH.flatXmap({
				try {
					val entries = ArrayList<Pair<ItemStack, Double>>()

					for (entry in getJson(it as Dynamic<Any?>).asJsonArray) {
						entry as JsonObject
						val item = entry["item"]?.asJsonObject ?: throw JsonParseException("Missing 'item'")
						val name = item["name"]?.asString ?: throw JsonParseException("Missing 'item.name'")
						val count = item["count"]?.asInt ?: throw JsonParseException("Missing 'item.count'")
						val chance = entry["chance"]?.asDouble ?: throw JsonParseException("Missing 'chance'")

						val itemObj = ForgeRegistries.ITEMS.getValue(ResourceLocation(name))
							?: throw JsonParseException("Unknown item $name")
						val stack = ItemStack(itemObj, count)

						entries.add(stack to chance)
					}

					DataResult.success(entries)
				} catch (err: JsonSyntaxException) {
					DataResult.error(err.message)
				} catch (err: JsonParseException) {
					DataResult.error(err.message)
				}
			}, {
				try {
					val listing = JsonArray()

					for ((item, chance) in it) {
						val entry = JsonObject()

						entry.add("item", JsonObject().also {
							it.addProperty("name", item.item.registryName!!.toString())
							it.addProperty("count", item.count)
						})

						entry.addProperty("chance", chance)

						listing.add(entry)
					}

					DataResult.success(Dynamic(JsonOps.INSTANCE, listing))
				} catch (err: JsonSyntaxException) {
					DataResult.error(err.message)
				} catch (err: JsonParseException) {
					DataResult.error(err.message)
				}
			})
		}

		val CODEC: Codec<LootTableBasicAppender> by lazy {
			RecordCodecBuilder.create {
				codecStart(it).and(
					ItemPairListCodec.fieldOf("entries").forGetter { appender -> appender.entries }
				).apply(it, ::LootTableBasicAppender)
			}
		}
	}
}
